function eventEmitter$1(valid_events) {
	let methods = {};
	let registered_handler = {};
	let event_handler_added_handler = null;
	let event_dispatched_handler = null;
	let event_handler_removed_handler = null;

	for (let event of valid_events) {
		registered_handler[event] = [];
	}

	const handlerInstalledForEvent = (event_name, fn) => {
		return registered_handler[event_name].filter(handler => {
			return Object.is(handler, fn)
		}).length > 0
	};

	methods.on = function on(event_name, fn) {
		return methods.addEventListener(event_name, fn)
	};

	methods.addEventListener = function addEventListener(event_name, fn) {
		if (!(event_name in registered_handler)) {
			throw new Error(`Invalid event '${event_name}'.`)
		} else if (handlerInstalledForEvent(event_name, fn)) {
			throw new Error(`This *exact* function was already registered for this event.`)
		}

		registered_handler[event_name].push(fn);

		if (typeof event_handler_added_handler === "function") {
			event_handler_added_handler(event_name, fn);
		}
	};

	methods.removeEventListener = function removeEventListener(event_name, fn) {
		if (!(event_name in registered_handler)) {
			throw new Error(`Invalid event '${event_name}'.`)
		} else if (!handlerInstalledForEvent(event_name, fn)) {
			throw new Error(`This function was not registered for this event.`)
		}

		registered_handler[event_name] = registered_handler[event_name].filter(handler => {
			return !Object.is(handler, fn)
		});

		if (typeof event_handler_removed_handler === "function") {
			event_handler_removed_handler(event_name, fn);
		}
	};

	return {
		setOnEventHandlerAddedHandler(fn) {
			event_handler_added_handler = fn;
		},

		setOnEventDispatchedHandler(fn) {
			event_dispatched_handler = fn;
		},

		setOnEventHandlerRemovedHandler(fn) {
			event_handler_removed_handler = fn;
		},

		install(obj) {
			for (const method of ["on", "addEventListener", "removeEventListener"]) {
				obj[method] = methods[method];
			}

			return function dispatchEvent(event_name, ...event_data) {
				if (!(event_name in registered_handler)) {
					throw new Error(`Invalid event '${event_name}'.`)
				}

				//
				// todo: maybe add "blocking" events
				// i.e. each event handler gets executed consecutively
				// example of code flow:
				// await handler()
				// await next_handler()
				// await all_handlers_dispatched()
				//
				for (const handler of registered_handler[event_name]) {
					setTimeout(handler, 0, ...event_data);
				}

				if (typeof event_dispatched_handler === "function") {
					setTimeout(event_dispatched_handler, 0, event_name, event_data);
				}
			}
		}
	}
}

function createPromise$1() {
	let promise, resolve, reject;

	promise = new Promise((a, b) => {
		resolve = a;
		reject = b;
	});

	return {promise, resolve, reject}
}

function createRandomIdentifier$1(len) {
	let str = "";
	let bytes = [];

	if (typeof crypto === "object" && typeof crypto.getRandomValues === "function") {
		bytes = new Uint8Array(len);

		crypto.getRandomValues(bytes);
	} else {
		for (let i = 0; i < len; ++i) {
			// will never exceed 255
			bytes.push(Math.floor(Math.random() * 256));
		}
	}

	for (let i = 0; i < len; ++i) {
		str += (bytes[i]).toString(16).slice(0, 1);
	}

	// sanity check output
	if (str.length !== len) {
		throw new Error(`createRandomIdentifier(${len}) failed for unknown reasons.`)
	}

	return str
}

function setupClient(id) {
	let client_instance = {
		id,
		current_stream: null,
		message_queue: [],
		public_interface: {
			id,
			sendMessage(msg) {
				// client is connected, send immediately
				if (client_instance.current_stream !== null) {
					client_instance.current_stream.write(
						`${JSON.stringify({
							message: msg
						})}\n`
					);
				} else {
					client_instance.message_queue.push(msg);
				}
			}
		}
	};

	const event_emitter = eventEmitter$1(["message", "disconnect"]);

	client_instance.dispatchEvent = event_emitter.install(client_instance.public_interface);

	return client_instance
}

function handleRequest(instance, relative_path, request, response) {
	if (relative_path === "create") {
		const client_id = createRandomIdentifier$1(32);

		instance.connected_clients.set(client_id, setupClient(client_id));

		response.write(JSON.stringify({
			id: client_id
		}) + "\n");
		response.end();

		return
	}
	else if (relative_path.startsWith("stream/")) {
		let client_id = relative_path.slice("stream/".length);

		if (!instance.connected_clients.has(client_id)) {
			throw new Error(`Invalid client identifier '${client_id}'.`)
		}

		const client_instance = instance.connected_clients.get(client_id);

		if (client_instance.current_stream !== null) {
			throw new Error(`Concurrent streams are not supported.`)
		}

		// release stream when socket closes
		request.socket.on("close", () => {
			client_instance.current_stream = null;

			instance.dispatchEvent("disconnect", client_instance.id);

			client_instance.dispatchEvent("disconnect");

			instance.connected_clients.delete(client_instance.id);
		});

		client_instance.current_stream = response;

		response.writeHead(200, {
			//"Content-Type": "text/event-stream",
			"Cache-Control": "no-cache",
			"Connection": "keep-alive"
		});

		instance.dispatchEvent("connect", client_instance.public_interface);

		client_instance.current_stream.write(
			`C813A843-5765-4DB3-8439-24AC1849117D\n`
		);

		// dump accumulated messages
		while (client_instance.message_queue.length) {
			const msg = client_instance.message_queue.shift();

			client_instance.current_stream.write(
				`${JSON.stringify({message: msg})}\n`
			);
		}

		return
	}
	else if (relative_path.startsWith("sendMessage/")) {
		let client_id = relative_path.slice("sendMessage/".length);

		if (!instance.connected_clients.has(client_id)) {
			throw new Error(`Invalid client identifier '${client_id}'.`)
		}

		const client_instance = instance.connected_clients.get(client_id);

		let body = "";

		request.on("data", chunk => body += chunk);

		request.on("end", () => {
			response.write("ok");
			response.end();

			const data = JSON.parse(body);

			instance.dispatchEvent("message", client_instance.id, data.message);
			client_instance.dispatchEvent("message", data.message);
		});

		return
	}

	throw new Error(`Unknown method '${relative_path}'.`)
}

async function index$1$1(http_port, base_url) {
	const {default: path} = await import('node:path');
	const {default: http} = await import('node:http');

	let instance = {
		ready_promise: createPromise$1(),
		connected_clients: new Map(),
		public_interface: {
			sendMessage(client_id, msg) {
				if (!instance.connected_clients.has(client_id)) {
					throw new Error(`No such client '${client_id}'.`)
				}

				const client_instance = instance.connected_clients.get(client_id);

				return client_instance.public_interface.sendMessage(msg)
			}
		}
	};

	const event_emitter = eventEmitter$1(["httpRequest", "connect", "message", "disconnect"]);

	instance.dispatchEvent = event_emitter.install(instance.public_interface);

	const normalized_base_url = path.normalize(base_url + "/");

	const server = http.createServer(async (request, response) => {
		const normalized_path = path.normalize(request.url);

		if (!normalized_path.startsWith(normalized_base_url)) {
			instance.dispatchEvent("httpRequest", {request, response});
		} else {
			const relative_path = normalized_path.slice(normalized_base_url.length);

			try {
				await handleRequest(instance, relative_path, request, response);
			} catch (error) {
				response.writeHead(500);
				response.write(`${error.message}`);
				response.end();
			}
		}
	});

	server.on("error", instance.ready_promise.reject);

	server.listen(http_port, () => {
		instance.public_interface.port = server.address().port;

		instance.ready_promise.resolve(instance.public_interface);
	});

	return instance.ready_promise.promise
}

//
// code taken from https://rob-blackbourn.medium.com/beyond-eventsource-streaming-fetch-with-readablestream-5765c7de21a1
// and adjusted to fit our use case
//
// Original Author: Rob Blackbourn, Date: Sep 23, 2019
//


function createLineDecoder() {
	return new TransformStream({
		start(controller) {
			controller.buf = "";
			controller.pos = 0;
		},

		transform(chunk, controller) {
			controller.buf += chunk;

			while (controller.pos < controller.buf.length) {
				if (controller.buf[controller.pos] == "\n") {
					const line = controller.buf.substring(0, controller.pos);
					controller.enqueue(line);
					controller.buf = controller.buf.substring(controller.pos + 1);
					controller.pos = 0;
				} else {
					controller.pos++;
				}
			}
		}
	})
}

function createWriteableEventStream(dispatchEvent) {
	return new WritableStream({
		start(controller) {
			dispatchEvent("start");
		},

		write(message, controller) {
			dispatchEvent("message", message);
		},

		close(controller) {
			dispatchEvent("close");
		},

		abort(reason) {
			dispatchEvent("close");
		}
	})
}

function SSEWithFetch(url, options = {}) {
	let instance = {
		public_interface: {}
	};

	const event_emitter = eventEmitter$1(["start", "message", "close", "error"]);

	const dispatchEvent = event_emitter.install(instance.public_interface);

	const line_decoder = createLineDecoder();
	const event_stream = createWriteableEventStream(dispatchEvent);

	fetch(url, options)
	.then(response => {
		return response.body
			.pipeThrough(new TextDecoderStream())
			.pipeThrough(line_decoder)
			.pipeTo(event_stream)
	})
	.catch(error => {
		dispatchEvent("error", error);
	});

	return instance.public_interface
}

async function createClient(endpoint_url) {
	const response = await fetch(`${endpoint_url}/create`);

	return (await response.json()).id
}

async function index$2(endpoint_url) {
	const ready_promise = createPromise$1();

	const id = await createClient(endpoint_url);

	let instance = {
		id,
		is_ready: false,
		message_queue: [],
		public_interface: {
			id,

			async sendMessage(msg) {
				await fetch(`${endpoint_url}/sendMessage/${id}`, {
					headers: {
						"Content-Type": "application/json"
					},
					method: "POST",
					body: JSON.stringify({message: msg})
				});
			}
		}
	};

	const event_emitter = eventEmitter$1(["message"]);

	instance.dispatchEvent = event_emitter.install(instance.public_interface);

	const event_source = SSEWithFetch(`${endpoint_url}/stream/${id}`);

	event_source.on("message", msg => {
		if (instance.is_ready === false) {
			if (msg === `C813A843-5765-4DB3-8439-24AC1849117D`) {
				instance.is_ready = true;

				ready_promise.resolve();

				setTimeout(() => {
					while (instance.message_queue.length) {
						const msg = instance.message_queue.shift();

						instance.dispatchEvent("message", msg);
					}
				}, 0);
			} else {
				instance.message_queue.push(JSON.parse(msg).message);
			}
		} else {
			instance.dispatchEvent("message", JSON.parse(msg).message);
		}
	});

	await ready_promise.promise;

	return instance.public_interface
}

function eventEmitter(valid_events) {
	let methods = {};
	let registered_handler = {};
	let event_handler_added_handler = null;
	let event_dispatched_handler = null;
	let event_handler_removed_handler = null;

	for (let event of valid_events) {
		registered_handler[event] = [];
	}

	const handlerInstalledForEvent = (event_name, fn) => {
		return registered_handler[event_name].filter(handler => {
			return Object.is(handler, fn)
		}).length > 0
	};

	methods.on = function on(event_name, fn) {
		return methods.addEventListener(event_name, fn)
	};

	methods.addEventListener = function addEventListener(event_name, fn) {
		if (!(event_name in registered_handler)) {
			throw new Error(`Invalid event '${event_name}'.`)
		} else if (handlerInstalledForEvent(event_name, fn)) {
			throw new Error(`This *exact* function was already registered for this event.`)
		}

		registered_handler[event_name].push(fn);

		if (typeof event_handler_added_handler === "function") {
			event_handler_added_handler(event_name, fn);
		}
	};

	methods.removeEventListener = function removeEventListener(event_name, fn) {
		if (!(event_name in registered_handler)) {
			throw new Error(`Invalid event '${event_name}'.`)
		} else if (!handlerInstalledForEvent(event_name, fn)) {
			throw new Error(`This function was not registered for this event.`)
		}

		registered_handler[event_name] = registered_handler[event_name].filter(handler => {
			return !Object.is(handler, fn)
		});

		if (typeof event_handler_removed_handler === "function") {
			event_handler_removed_handler(event_name, fn);
		}
	};

	return {
		setOnEventHandlerAddedHandler(fn) {
			event_handler_added_handler = fn;
		},

		setOnEventDispatchedHandler(fn) {
			event_dispatched_handler = fn;
		},

		setOnEventHandlerRemovedHandler(fn) {
			event_handler_removed_handler = fn;
		},

		install(obj) {
			for (const method of ["on", "addEventListener", "removeEventListener"]) {
				obj[method] = methods[method];
			}

			return function dispatchEvent(event_name, ...event_data) {
				if (!(event_name in registered_handler)) {
					throw new Error(`Invalid event '${event_name}'.`)
				}

				//
				// todo: maybe add "blocking" events
				// i.e. each event handler gets executed consecutively
				// example of code flow:
				// await handler()
				// await next_handler()
				// await all_handlers_dispatched()
				//
				for (const handler of registered_handler[event_name]) {
					setTimeout(handler, 0, ...event_data);
				}

				if (typeof event_dispatched_handler === "function") {
					setTimeout(event_dispatched_handler, 0, event_name, event_data);
				}
			}
		}
	}
}

function createRandomIdentifier(len) {
	let str = "";
	let bytes = [];

	if (typeof crypto === "object" && typeof crypto.getRandomValues === "function") {
		bytes = new Uint8Array(len);

		crypto.getRandomValues(bytes);
	} else {
		for (let i = 0; i < len; ++i) {
			// will never exceed 255
			bytes.push(Math.floor(Math.random() * 256));
		}
	}

	for (let i = 0; i < len; ++i) {
		str += (bytes[i]).toString(16).slice(0, 1);
	}

	// sanity check output
	if (str.length !== len) {
		throw new Error(`createRandomIdentifier(${len}) failed for unknown reasons.`)
	}

	return str
}

function createPromise() {
	let promise, resolve, reject;

	promise = new Promise((a, b) => {
		resolve = a;
		reject = b;
	});

	return {promise, resolve, reject}
}

function logarithmic(base_value, index) {
	return Math.log(base_value) * ((index + 1) * 10)
}

function retransmitTimerValuesGenerator(max_attempts, {
	max_timeout = 10000,
	base_value = 250,
	calculate_value_in_ms = logarithmic
} = {}) {

	let num_attempts_left = max_attempts;
	let num_attempts = 0;

	return {
		//
		// returns the next calculated timeout value that should be used
		//
		getNextTimeoutValue() {
			if (0 > (num_attempts_left - 1)) {
				throw new Error(
					`Called getNextTimeoutValue() after reaching the maximum number of attempts (${num_attempts}/${max_attempts}).`
				)
			}

			let timeout = Math.floor(calculate_value_in_ms(base_value, num_attempts));

			if (timeout > max_timeout) timeout = max_timeout;

			++num_attempts;
			--num_attempts_left;

			return timeout
		},

		getNumberOfAttemptsTaken() {
			return num_attempts
		},

		getNumberOfAttemptsLeft() {
			return num_attempts_left
		}
	}
}

retransmitTimerValuesGenerator.fromTimeBudget = function(time_budget, {
	base_value = 250,
	calculate_value_in_ms = logarithmic
} = {}) {
	let timer = retransmitTimerValuesGenerator(Infinity, {
		max_timeout: Infinity,
		base_value,
		calculate_value_in_ms
	});

	let accumulated_time = 0, calculated_number_of_attempts = 0;

	while (true) {
		let current_value = timer.getNextTimeoutValue();

		if ((accumulated_time + current_value) > time_budget) break

		accumulated_time += current_value;
		++calculated_number_of_attempts;
	}

	//
	// sanity check
	//
	if (accumulated_time > time_budget) {
		throw new Error(`Something went wrong, accumulated_time is bigger than time_budget!`)
	}

	return retransmitTimerValuesGenerator(calculated_number_of_attempts, {
		max_timeout: Infinity,
		base_value,
		calculate_value_in_ms
	})
};

retransmitTimerValuesGenerator.algorithms = {logarithmic};

async function sendRequestWithOptions(instance, options, request_data) {
	instance.assertReadyAndNotClosed();

	const {timeout, max_retransmissions} = options;
	const request_id = createRandomIdentifier(12);
	const request_promise = createPromise();

	instance.trace(
		`i will be sending message '${JSON.stringify(request_data)}' with request_id = '${request_id}', ` +
		`timeout = '${timeout}ms' and max_retransmissions = '${max_retransmissions}'`
	);

	// keep track of timer to clear the timeout when response arrives
	let timeout_timer = null;

	const transmitRequestInformation = async (attempt_no) => {
		const max_attempts = (max_retransmissions === Infinity ? -1 : max_retransmissions);

		await instance.sendJSONData({
			cmd: "request",
			request_id,
			data: request_data,
			attempt_no,
			max_attempts
		});
	};

	if (timeout > 0) {
		const generator = retransmitTimerValuesGenerator(max_retransmissions);

		const cancelRequest = (reason) => {
			const {reject} = request_promise;

			setTimeout(reject, 0, new Error(reason));

			instance.open_requests.delete(request_id);
		};

		const timer_handler = (is_first_call = false) => {
			// it's an error if the request does not exist anymore
			if (!instance.open_requests.has(request_id)) {
				throw new Error(`Detected request (${request_id}) that was not properly disposed. This is a bug.`)
			}

			const request_object = instance.open_requests.get(request_id);

			// 'is_first_call' indicates that the timer handler was
			// called the first time which means the request
			// time'd out

			const attempts_left = generator.getNumberOfAttemptsLeft();

			if (is_first_call) {
				// now we need to decide whether to reject the
				// request or retransmit it
				instance.debug(`request (${request_id}) has reached timeout of ${timeout}ms.`);

				if (!attempts_left) {
					return cancelRequest(`Timeout of '${timeout}ms' reached.`)
				}

				return timer_handler(false)
			}

			if (!instance.open_requests.has(request_id)) {
				instance.debug(`request (${request_id}) has been settled. No need for further retransmission.`);

				return
			}

			if (attempts_left > 0) {
				const delay = generator.getNextTimeoutValue();
				const attempts_left_str = (attempts_left === Infinity) ? "*unlimited*" : attempts_left;

				instance.trace(
					`we have ${attempts_left_str} more attempts left to settle the request '${request_id}' ; ` +
					`will retransmit request in ${delay}ms.`
				);

				request_object.timer = setTimeout(async () => {
					await transmitRequestInformation(generator.getNumberOfAttemptsTaken());
					await timer_handler();
				}, delay, false);

				return
			}

			return cancelRequest(
				`Reached maximum amount of retransmission attempts ${max_retransmissions}.`
			)
		};

		timeout_timer = setTimeout(timer_handler, timeout, true);
	}

	instance.open_requests.set(request_id, {
		request_promise,
		timer: timeout_timer
	});

	await transmitRequestInformation(0);

	return request_promise.promise
}

function pruneJavaScriptMapObject(map, max_size = 1000) {
	if (max_size > map.size) {
		return
	}

	const map_keys = Array.from(map.keys());
	const keys_to_be_removed = map_keys.slice(
		0, map.size - max_size
	);

	for (const key of keys_to_be_removed) {
		map.delete(key);
	}
}

function pruneRequestsCache(instance) {
	pruneJavaScriptMapObject(instance.handled_requests);
}

async function handleIncomingRequest(instance, message) {
	let response = null, from_cache = false, exec_time = -1;

	if (instance.handled_requests.has(message.request_id)) {
		response = instance.handled_requests.get(message.request_id);

		from_cache = true;
	} else if (instance.pending_responses.has(message.request_id)) {
		instance.trace(
			`received a request (${message.request_id}) that is still being processed.`
		);

		return
	} else {
		instance.pending_responses.set(message.request_id, 1);

		response = {
			error: true,
			data: null
		};

		try {
			let exec_time_start = performance.now();

			response.data = await instance.public_interface.requestHandler(message.data, null, {
				debug: {
					instance,
					message
				}
			});
			response.error = false;

			if (instance.debugging_mode === true) {
				exec_time = performance.now() - exec_time_start;
			}
		} catch (error) {
			let error_message = "Uncaught error in remote requestHandler.";

			if (instance.debugging_mode === true) {
				error_message += "\n";
				error_message += `Message: ${error.message}\n`;
				error_message += "\n";
				error_message += error.stack;
			}

			response.data = error_message;
		}

		instance.handled_requests.set(message.request_id, response);
		instance.pending_responses.delete(message.request_id);
	}

	let from_cache_str = from_cache ? " (from cache)" : "";

	instance.trace(`for the message '${JSON.stringify(message)}' my response is '${JSON.stringify(response)}'${from_cache_str}`);

	instance.sendJSONData({
		cmd: "response",
		original_request_id: message.request_id,
		response,
		from_cache,
		exec_time
	});

	pruneRequestsCache(instance);
}

async function handleIncomingResponse(instance, message) {
	const {original_request_id} = message;

	if (!instance.open_requests.has(message.original_request_id)) {
		instance.debug(`no pending request with id '${original_request_id}'`);

		return
	}

	const open_request = instance.open_requests.get(original_request_id);

	if (open_request.timer !== null) {
		clearTimeout(open_request.timer);
	}

	const {resolve, reject} = open_request.request_promise;

	if (message.response.error) {
		setTimeout(reject, 0, new Error(message.response.data));
	} else {
		setTimeout(resolve, 0, message.response.data);
	}

	instance.open_requests.delete(original_request_id);
}

async function onMessageReceived(instance, message) {
	instance.trace(`i will be handling the following message '${JSON.stringify(message)}'`);

	if (message.cmd === "request") {
		await handleIncomingRequest(instance, message);
	} else if (message.cmd === "response") {
		await handleIncomingResponse(instance, message);
	} else {
		instance.debug(`bad message`, message);
	}
}

function createRequestResponseProtocol(api, label = "") {
	const connection_id = createRandomIdentifier(12);
	const synchronize_token = createRandomIdentifier(32);
	let synchronized_promise = createPromise();

	//
	// Wrap api.sendMessage to ignore errors that are thrown
	// by that function.
	//
	const sendMessageViaAPI = async (instance, message) => {
		try {
			await api.sendMessage(message);
		} catch (e) {
			instance.debug(`api.sendMessage failed with '${e.message}'`);
		}
	};

	let instance = {
		connection_id,
		label,

		open_requests: new Map(),
		//
		// Keep a map of handled requests so if the same
		// request is received a second time, the result is re-used
		// and the request handler isn't called a second time.
		//
		handled_requests: new Map(),
		//
		// Keep track of pending responses so we don't execute
		// the request handler a second time.
		//
		pending_responses: new Map(),
		ready: false,
		closed: false,
		debugging_mode: false,

		//
		// the default timeout for a request
		// this timeout applies to one-shot requests
		// and regular requests with zero retransmission attempts.
		//
		default_request_timeout: 500,

		//
		// the default max. retransmissions for a request
		//
		default_max_retransmissions: 10,

		debug(...args) {
			if (instance.public_interface.debug !== true) return

			console.log(`[${label} (D)]`, ...args);
		},

		trace(...args) {
			if (instance.public_interface.trace !== true) return

			console.log(`[${label} (T)]`, ...args);
		},

		sendJSONData(data) {
			sendMessageViaAPI(instance, `start{` + JSON.stringify(data) + `}end`);
		},

		messageHandler(message) {
			if (message.startsWith(`@anio-js-foundation/requestResponseProtocol:`)) {
				message = message.slice(`@anio-js-foundation/requestResponseProtocol:`.length);

				if (message.startsWith("sync:")) {
					const sync_token = message.slice("sync:".length);

					sendMessageViaAPI(instance, `@anio-js-foundation/requestResponseProtocol:ack:${sync_token}`);
				} else if (message.startsWith("ack:")) {
					const sync_token = message.slice("ack:".length);

					if (sync_token === synchronize_token && !instance.ready) {
						instance.ready = true;

						synchronized_promise.resolve();
					}
				} else {
					instance.debug("bad message", message);
				}
			} else if (instance.ready && message.startsWith("start{") && message.endsWith("}end")) {
				message = message.slice("start{".length);
				message = message.slice(0, message.length - "}end".length);

				onMessageReceived(instance, JSON.parse(message));
			} else {
				instance.debug("bad message", message);
			}
		},

		assertReadyAndNotClosed() {
			if (!instance.ready) {
				throw new Error(`Cannot send request when protocol is not ready.`)
			} else if (instance.closed) {
				throw new Error(`Cannot send request when protocol is closed.`)
			}
		},

		public_interface: {
			debug: false,
			trace: false,

			connection_id,

			_setDebuggingMode(value) {
				instance.debugging_mode = value;
			},

			ready() {
				return synchronized_promise.promise
			},

			sendRequestWithOptions({
				timeout = instance.default_request_timeout,
				max_retransmissions = instance.default_max_retransmissions
			} = {}, request_data) {
				return sendRequestWithOptions(instance, {timeout, max_retransmissions}, request_data)
			},

			setDefaultMaxRetransmissions(value) {
				instance.default_max_retransmissions = value;
			},

			getDefaultMaxRetransmissions(value) {
				return instance.default_max_retransmissions
			},

			setDefaultRequestTimeout(value) {
				instance.default_request_timeout = value;
			},

			getDefaultRequestTimeout() {
				return instance.default_request_timeout
			},

			requestHandler(request) {
				return `Please set instance.requestHandler to handle requests. Incoming request data was: ${JSON.stringify(request)}.`
			},

			sendRequest(request_data) {
				return instance.public_interface.sendRequestWithOptions({}, request_data)
			},

			/* special api
			withTimeout(with_timeout_value) {
				return {
					sendRequest(request_data) {
						return instance.public_interface.sendRequestWithOptions({
							timeout: with_timeout_value
						}, request_data)
					},

					withMaxRetransmissions(with_max_retransmissions_value) {
						return {
							sendRequest(request_data) {
								return instance.public_interface.sendRequestWithOptions({
									timeout: with_timeout_value,
									max_retransmissions: with_max_retransmissions_value
								}, request_data)
							}
						}
					}
				}
			},

			withMaxRetransmissions(with_max_retransmissions_value) {
				return {
					sendRequest(request_data) {
						return instance.public_interface.sendRequestWithOptions({
							max_retransmissions: with_max_retransmissions_value
						})
					},

					withTimeout(with_timeout_value) {
						return {
							sendRequest(request_data) {
								return instance.public_interface.sendRequestWithOptions({
									timeout: with_timeout_value,
									max_retransmissions: with_max_retransmissions_value
								}, request_data)
							}
						}
					}
				}
			},
			special api */

			closeAllPendingRequests(reason = `closeAllPendingRequests() was called.`) {
				for (const [request_id, open_request] of instance.open_requests) {
					const {reject} = open_request.request_promise;

					setTimeout(reject, 0, new Error(reason));

					// make sure to cancel timer
					if (open_request.timer !== null) {
						clearTimeout(open_request.timer);
					}

					instance.open_requests.delete(request_id);
				}
			},

			close() {
				if (instance.closed) return

				instance.closed = true;

				setTimeout(
					instance.public_interface.closeAllPendingRequests, 0, "close() was called."
				);

				api.removeEventListener("message", instance.messageHandler);
			}
		}
	};

	api.on("message", instance.messageHandler);

	//
	// Slave can miss this message so repeat it
	// every once in a while to get the master and slave
	// synced up.
	//
	const delay_map = [50, 50, 50, 100, 100, 150, 150, 150, 250, 500, 750, 1000];
	let delay_index = 0;

	const synchronize = async () => {
		if (instance.closed) return
		if (instance.ready) return

		instance.debug("attempt to synchronize");

		await sendMessageViaAPI(instance, `@anio-js-foundation/requestResponseProtocol:sync:${synchronize_token}`);

		// by default use the last value from the delay map
		let amount = delay_map[delay_map.length - 1];

		if (delay_map.length > delay_index) {
			amount = delay_map[delay_index];

			++delay_index;
		}

		instance.debug(`will be calling synchronize() in ${amount}ms`);

		setTimeout(synchronize, amount);
	};

	setTimeout(synchronize, 0);

	return instance.public_interface
}

async function index$1(port, base_url, onHTTPResourceRequest = () => {}) {
	let instance = {
		connected_clients: new Map(),
		client_mapping: new Map(),

		public_interface: {
			getClients() {
				return Array.from(instance.connected_clients.keys())
			},

			getClientById(client_id) {
				if (!instance.connected_clients.has(client_id)) {
					throw new Error(`Unknown client '${client_id}'.`)
				}

				return instance.connected_clients.get(client_id)
			}
		}
	};

	const server = await index$1$1(port, base_url);

	const event_emitter = eventEmitter(["connect", "disconnect"]);

	const dispatchEvent = event_emitter.install(instance.public_interface);

	instance.dispatchEvent = dispatchEvent;

	server.on("httpRequest", ({request, response}) => {
		onHTTPResourceRequest(request, response);
	});

	server.on("connect", client => {
		const protocol = createRequestResponseProtocol(client, `Server -> Client ${client.id.slice(0, 4)}`);

		protocol.ready().then(() => {
			instance.connected_clients.set(protocol.connection_id, protocol);
			instance.client_mapping.set(client.id, protocol);

			instance.dispatchEvent("connect", protocol);
		});
	});

	server.on("disconnect", (client_id) => {
		// client never got past protocol sync phase
		if (!instance.client_mapping.has(client_id)) {
			return
		}

		const protocol = instance.client_mapping.get(client_id);

		instance.dispatchEvent("disconnect", {
			connection_id: protocol.connection_id
		});

		instance.connected_clients.delete(protocol.connection_id);
		instance.client_mapping.delete(client_id);
	});

	instance.public_interface.port = server.port;

	return instance.public_interface
}

async function index(endpoint, request_handler) {
	const client = await index$2(endpoint);

	const protocol = createRequestResponseProtocol(client, `Client ${client.id.slice(0, 7)} -> Server`);

	protocol.requestHandler = request_handler;

	await protocol.ready();

	return protocol
}

export { index as createClient, index$1 as createServer };
